
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:love/love.dart';
import 'package:provider/provider.dart';

class Store<State, Event> {
  Store(this.state, this.dispatch);

  final State state;
  final Dispatch<Event> dispatch;
}

class StoreProvider<State, Event> extends DeferredInheritedProvider<EffectSystem<State, Event>, Store<State, Event>> {

  StoreProvider({
    Key? key,
    bool? lazy,
    required Create<EffectSystem<State, Event>> create,
    UpdateShouldNotify<Store<State, Event>>? updateShouldNotify,
    TransitionBuilder? builder,
    Widget? child,
  }): super(
    key: key,
    lazy: lazy,
    create: create,
    updateShouldNotify: updateShouldNotify,
    builder: builder,
    child: child,
    startListening: _systemStartListening(),
  );

  StoreProvider.value({
    Key? key,
    bool? lazy,
    required EffectSystem<State, Event> value,
    UpdateShouldNotify<Store<State, Event>>? updateShouldNotify,
    TransitionBuilder? builder,
    Widget? child,
  }): super.value(
    key: key,
    lazy: lazy,
    value: value,
    updateShouldNotify: updateShouldNotify,
    builder: builder,
    child: child,
    startListening: _systemStartListening(),
  );
}


DeferredStartListening<EffectSystem<State, Event>, Store<State, Event>> _systemStartListening<State, Event>() {
  return (context, setState, system, __) {
    bool isDisposed = false;
    final dispose = system.run(
      effect: (state, oldState, event, dispatch) {
        if (isDisposed) return; 
        setState(Store(state, dispatch));
      },
    );
    return () {
      if (isDisposed) return;
      isDisposed = true;
      dispose();
    };
  };
}